#include "StdAfx.h"
#include "MeshWarp.h"

CMeshWarp::CMeshWarp(void)
{
}

CMeshWarp::~CMeshWarp(void)
{
}

void CMeshWarp::FantsResamplingHeight(CImage *src, CImage *dest, int srcStart, int srcEnd,
									  int destStart, int destEnd, int w)
{
	int outlen = destEnd - destStart;
	int inlen = srcEnd - srcStart;
	float scale = (float)outlen / (float)inlen;
	float insfac = 1.0f / scale;
	float outseg = insfac;
	float inseg = 1.0f;
	float accumulatorR, accumulatorG, accumulatorB;
	int x = destStart;
	int u = srcStart;
	float intensityR, intensityG, intensityB;
	COLORREF color1, color2;

	accumulatorR = accumulatorG = accumulatorB = 0.0f;
	while (x <= destEnd && u <= srcEnd) {
		if (u < src->GetHeight() - 1) {
			color1 = GetPointColor(src, w, u);
			color2 = GetPointColor(src, w, u+1);
			ComputeCount++;
			ComputeCount++;
		}
		intensityR = (inseg * GetRValue(color1)) + ((1 - inseg) * GetRValue(color2));
		intensityG = (inseg * GetGValue(color1)) + ((1 - inseg) * GetGValue(color2));
		intensityB = (inseg * GetBValue(color1)) + ((1 - inseg) * GetBValue(color2));
		
		if (inseg < outseg)
		{
			accumulatorR += (intensityR * inseg);
			accumulatorG += (intensityG * inseg);
			accumulatorB += (intensityB * inseg);
			outseg -= inseg;
			inseg = 1.0f;
			u++;
		} 
		else
		{
			accumulatorR += (intensityR * outseg);
			accumulatorG += (intensityG * outseg);
			accumulatorB += (intensityB * outseg);

			//  Ʈ °.
			// SetPixel ʹ . ̰  M$Ǽ, PointColor 
			PointColor(dest, w, x, RGB((float)accumulatorR * scale, (float)accumulatorG * scale, (float)accumulatorB * scale));
			accumulatorR = accumulatorG = accumulatorB = 0.0f;
			inseg -= outseg;
			outseg = insfac;
			x++;

			ComputeCount++;
		}
		ComputeCount+=10;
	}
}

void CMeshWarp::FantsResamplingWidth(CImage *src, CImage *dest, int srcStart, int srcEnd,
		int destStart, int destEnd, int h)
{
	int outlen = destEnd - destStart;
	int inlen = srcEnd - srcStart;
	float scale = (float)outlen / (float)inlen;
	float insfac = 1.0f / scale;
	float outseg = insfac;
	float inseg = 1.0f;
	float accumulatorR, accumulatorG, accumulatorB;
	int x = destStart;
	int u = srcStart;
	float intensityR, intensityG, intensityB;
	COLORREF color1, color2;

	accumulatorR = accumulatorG = accumulatorB = 0.0f;
	while (x <= destEnd && u <= srcEnd) {
		if (u < src->GetWidth() - 1) {
			color1 = GetPointColor(src, u, h);
			color2 = GetPointColor(src, u+1, h);
			ComputeCount++;
			ComputeCount++;
		}
		intensityR = (inseg * GetRValue(color1)) + ((1 - inseg) * GetRValue(color2));
		intensityG = (inseg * GetGValue(color1)) + ((1 - inseg) * GetGValue(color2));
		intensityB = (inseg * GetBValue(color1)) + ((1 - inseg) * GetBValue(color2));

		if (inseg < outseg) {
			accumulatorR += (intensityR * inseg);
			accumulatorG += (intensityG * inseg);
			accumulatorB += (intensityB * inseg);
			outseg -= inseg;
			inseg = 1.0f;
			u++;
		} else {
			accumulatorR += (intensityR * outseg);
			accumulatorG += (intensityG * outseg);
			accumulatorB += (intensityB * outseg);
			
			//  Ʈ °.
			// SetPixel ʹ . ̰  M$Ǽ, PointColor 
			PointColor(dest, x, h, RGB((float)accumulatorR * scale, (float)accumulatorG * scale, (float)accumulatorB * scale));
			accumulatorR = accumulatorG = accumulatorB = 0.0f;
			inseg -= outseg;
			outseg = insfac;
			x++;

			ComputeCount++;
			ComputeCount++;
		}
		ComputeCount+=9;
	}
}

void CMeshWarp::MeshWarp(CImage *src, CImage *dest, MESHPOINT *srcMesh, MESHPOINT *destMesh,
		int widthMesh, int heightMesh)
{
	m_nWidth = src->GetWidth();
	m_nHeight = src->GetHeight();
	m_nBitDepth = src->GetBPP();
	int x, y, inx, iny;
	int **lineBuffer = NULL;
	int **srcBuffer = NULL;
	MESHPOINT *tempMesh = NULL;

	/* New canvas */
	dest->Create(m_nWidth, m_nHeight, m_nBitDepth);
	
	/****************************
		ONE PASS : 켱  ְ õѴ.
	****************************/

	/* Create Linear function : Each row */

	/* Temporary Dest Mesh */
	tempMesh = (MESHPOINT *)malloc(sizeof(MESHPOINT) * widthMesh * heightMesh);
	for (x = 0; x < widthMesh * heightMesh; x++) {
		tempMesh[x].x = destMesh[x].x;
		tempMesh[x].y = srcMesh[x].y;
		ComputeCount++;
	}

	/* Allocate Linear function */
	lineBuffer = (int **)malloc(sizeof(int *) * widthMesh);
	srcBuffer = (int **)malloc(sizeof(int *) * widthMesh);
	for (x = 0; x < widthMesh; x++) {
		lineBuffer[x] = (int *)malloc(sizeof(int) * m_nHeight);
		srcBuffer[x] = (int *)malloc(sizeof(int) * m_nHeight);
		ComputeCount++;
	}

	for (x = 0; x < widthMesh; x++)
		for (y = 0; y < heightMesh - 1; y++) {
			int c = y * widthMesh + x;		/* Current Mesh Point */
			int b = c + widthMesh;			/* Bottom Mesh Point */

			/* Compute linear rate */
			float rate, srcrate;
			if (tempMesh[b].y - tempMesh[c].y != 0)
				rate = (float)(tempMesh[b].x - tempMesh[c].x) / (float)(tempMesh[b].y - tempMesh[c].y);
			else
				rate = 0.0f;
			if (srcMesh[b].y - srcMesh[c].y != 0)
				srcrate = (float)(srcMesh[b].x - srcMesh[c].x) / (float)(srcMesh[b].y - srcMesh[c].y);
			else
				srcrate = 0.0f;

			ComputeCount+=4;

			/* Compute linear buffer */
			for (iny = srcMesh[c].y; iny <= srcMesh[b].y; iny++) {
				lineBuffer[x][iny] = (int)((float)tempMesh[c].x + rate * ((float)iny - tempMesh[c].y));
				srcBuffer[x][iny] = (int)((float)srcMesh[c].x + srcrate * ((float)iny - srcMesh[c].y));
				ComputeCount++;
			}
		}

	// Fan's Resampling : one pass (Each row) //
	for (x = 0; x < widthMesh - 1; x++)
		for (y = 0; y < heightMesh - 1; y++) {
			int lt = y * widthMesh + x;		// lt = left top
			int rt = lt + 1;				// rt = right top
			int lb = lt + widthMesh;		// lb = left bottom
			int rb = lb + 1;				// rb = right bottom
			for (iny = srcMesh[lt].y; iny <= srcMesh[lb].y; iny++)
			{
				FantsResamplingWidth(src, 
									 dest, 
									 srcBuffer[x][iny],			//srcMesh[lt].x,					// Source Start //
									 srcBuffer[x + 1][iny],		//srcMesh[rt].x,					// Source End // 
									 lineBuffer[x][iny],		// Dest Start //
									 lineBuffer[x + 1][iny],	// Dest End //
									 iny);						// Height //

				ComputeCount++;
			}
			ComputeCount+=4;
		}

	// Memory Deallocation //
	for (x = 0; x < widthMesh; x++) {
		free(srcBuffer[x]);
		free(lineBuffer[x]);
		ComputeCount++;
	}
	free(srcBuffer);
	free(lineBuffer);

	//*src = *dest;
	dest->BitBlt(src->GetDC(),0,0);
	src->ReleaseDC();
	ComputeCount+=3;
	/****************************
		TWO PASS :   ְ ǽѴ.
	****************************/

	// Create Linear function : Each col 

	// Temporary Dest Mesh 
	for (x = 0; x < widthMesh * heightMesh; x++) {
		tempMesh[x].x = srcMesh[x].x;
		tempMesh[x].y = destMesh[x].y;
		ComputeCount++;
	}
	// Allocate Linear function 
	lineBuffer = (int **)malloc(sizeof(int *) * heightMesh);
	srcBuffer = (int **)malloc(sizeof(int *) * heightMesh);
	for (x = 0; x < heightMesh; x++) {
		lineBuffer[x] = (int *)malloc(sizeof(int) * m_nWidth);
		srcBuffer[x] = (int *)malloc(sizeof(int) * m_nWidth);
		ComputeCount++;
	}

	for (y = 0; y < heightMesh; y++)
		for (x = 0; x < widthMesh - 1; x++) {
			int c = y * widthMesh + x;		// Current Mesh Point 
			int r = c + 1;					// Right Mesh Point 

			// Compute linear rate 
			float rate, srcrate;
			if (tempMesh[r].x - tempMesh[c].x != 0)
				rate = (float)(tempMesh[r].y - tempMesh[c].y) / (float)(tempMesh[r].x - tempMesh[c].x);
			else
				rate = 0.0f;

			if (srcMesh[r].x - srcMesh[c].x != 0)
				srcrate = (float)(srcMesh[r].y - srcMesh[c].y) / (float)(srcMesh[r].x - srcMesh[c].x);
			else
				srcrate = 0.0f;
			
			ComputeCount+=4;
			// Compute linear buffer 
			for (inx = srcMesh[c].x; inx <= srcMesh[r].x; inx++) {
				lineBuffer[y][inx] = (int)((float)tempMesh[c].y + rate * ((float)inx - tempMesh[c].x));
				srcBuffer[y][inx] = (int)((float)srcMesh[c].y + srcrate * ((float)inx - srcMesh[c].x));
				ComputeCount+=2;
			}
		}

	// Fan's Resampling : two pass (Each col) 
	for (x = 0; x < widthMesh - 1; x++)
		for (y = 0; y < heightMesh - 1; y++) {
			int lt = y * widthMesh + x;
			int rt = lt + 1;
			int lb = lt + widthMesh;
			int rb = lb + 1;
			for (inx = srcMesh[lt].x; inx <= srcMesh[rt].x; inx++)
			{
				FantsResamplingHeight(src, 
									  dest, 
									  srcBuffer[y][inx],			//srcMesh[lt].y,					// Source Start 
									  srcBuffer[y + 1][inx],		//srcMesh[lb].y,					// Source End  
									  lineBuffer[y][inx],			// Dest Start 
									  lineBuffer[y + 1][inx],		// Dest End 
									  inx);							// Height 
				ComputeCount++;
			}
			ComputeCount+=4;
		}

	for (x = 0; x < heightMesh; x++) {
		free(srcBuffer[x]);
		free(lineBuffer[x]);
		ComputeCount++;
	}
	free(srcBuffer);
	free(lineBuffer);
	free(tempMesh);
}

COLORREF CMeshWarp::FindColorMedian(BYTE *r, BYTE *g, BYTE *b, int wsize)
{
	unsigned int		vec_sum, min_sum = 10000;
	int					index = 0;
	for (int i = 0; i < wsize; i++) {
		vec_sum = 0;
		for (int j = 0; j < wsize; j++) {
			vec_sum += abs(r[i] - r[j]);
			vec_sum += abs(g[i] - g[j]);
			vec_sum += abs(b[i] - b[j]);
			ComputeCount++;
		}
		if (vec_sum <= min_sum) {
			min_sum = vec_sum;
			index = i;
			ComputeCount++;
		}
	}
	return RGB(r[index], g[index], b[index]);
}

// ְ Ƣ Ʈ ִ հ ʴ. (伥  ȿ.)
void CMeshWarp::MedianFilter(CImage *src, CImage *dest, int kwidth)
{
	int				x, y, i, j;
	int				nWidth = src->GetWidth();
	int				nHeight = src->GetHeight();
	int				nkHeight = kwidth, nkWidth = kwidth;
	int				kerx, kery, index, xextra, yextra;
	BYTE			windowR[9], windowG[9], windowB[9];		/* storage for window samples */
	int				xpad, ypad;
	COLORREF		color;
	nkHeight = nkWidth = 3;

	yextra = (nkHeight / 2) * 2;
	ypad = yextra / 2;
	xextra = (nkWidth / 2) * 2;
	xpad = xextra / 2;

	ComputeCount+=5;
	for (y = ypad; y < (nHeight - ypad); y++) {
		for (x = xpad; x < (nWidth - xpad); x++) {
			index = 0;
			for (kery = -1 * ypad; kery <= ypad; kery++)
				for (kerx = - 1 * xpad; kerx <= xpad; kerx++) {
					color = GetPointColor(src, x + kerx, y + kery);

					windowR[index] = GetRValue(color);
					windowG[index] = GetGValue(color);
					windowB[index++] = GetBValue(color);
					ComputeCount++;
				}
			PointColor(dest, x, y, FindColorMedian(windowR, windowG, windowB, kwidth * kwidth));
		}
		if (y == ypad) {
			for (i = 0; i < ypad; i++) {
				PointColor(dest, 0, i, GetPointColor(dest, xpad, ypad));
				for (j = xpad; j < nWidth - xpad; j++)
				{
					PointColor(dest, j, i, GetPointColor(dest, j, ypad));
					ComputeCount++;
				}
				PointColor(dest, nWidth - xpad, i, GetPointColor(dest, xextra, ypad));
				ComputeCount++;
			}
		} else if (y == nHeight - yextra) {
			for (i = 0; i < ypad; i++) {
				PointColor(dest, 0, nHeight - yextra + 1 + i, GetPointColor(dest ,xpad, nHeight - yextra));
				for (j = xpad; j < nWidth - xpad; j++)
				{
					PointColor(dest, j, nHeight - yextra + 1 + i, GetPointColor(dest, j, nHeight - yextra));
					ComputeCount++;
				}
				PointColor(dest, nWidth - xpad, nHeight - yextra + 1 + i, GetPointColor(dest, xextra, nHeight - yextra));
				ComputeCount++;
			}
		} else {
			PointColor(dest, 0, y, GetPointColor(dest, xpad, y));
			PointColor(dest,nWidth - xpad, y, GetPointColor(dest, xextra, y));
			ComputeCount++;
			ComputeCount++;
		}
	}
}